@@ -3,10 +3,14 @@
 mod args;
 mod cd;
 mod cp_mv;
+mod exit;
+mod mkdir;
 mod rm;
 mod sleep;
 
 pub use cd::*;
 pub use cp_mv::*;
+pub use exit::*;
+pub use mkdir::*;
 pub use rm::*;
 pub use sleep::*;
